import pytest

from arthurai.core.models import ArthurModel
from arthurai.core.alerts import Metric, MetricType
from arthurai.client.base import BaseApiClient
from arthurai.common.exceptions import ArthurUserError
from tests import MockResponse

metric_id = "4e18df3d-307d-4b3f-aba4-a434ca893c8a"

metric_dict = {
    "name": "Custom Metric A",
    "endpoint": "/api/v3/models/4e18df3d-307d-4b3f-aba4-a434ca893c8a/inferences/query",
    "query": {
        "select": [
            {
                "function": "rate",
                "parameters": {
                    "property": "class_a",
                    "comparater": "gte",
                    "value": 0.5
                }
            }
        ],
        "filter": [
            {
                "property": "city",
                "comparator": "eq",
                "value": "Boston"
            }
        ]
    },
    "type": "model_performance_metric"
}


def patch_metric_post_request(*args, **kwargs):
    metric_to_create = args[2]
    assert "name" in metric_to_create
    assert "query" in metric_to_create
    assert "endpoint" in metric_to_create
    assert "https://test.ai" not in metric_to_create["endpoint"]
    assert "/inferences/query" in metric_to_create["endpoint"]
    assert metric_to_create["name"] == "Custom Metric A"
    metric_dict["name"] = metric_to_create["name"]
    metric_dict["query"] = metric_to_create["query"]
    metric_dict["endpoint"] = metric_to_create["endpoint"]
    metric_dict["id"] = metric_id
    return MockResponse(metric_dict, 201)


def patch_metric_get_request(*args, **kwargs):
    assert args[1] == f'None/models/4a57c553-e787-4307-88f1-3747ec9130f5/metrics'
    assert kwargs["params"]["page"] == str(TestMetrics.current_page_num)
    metric_list = {
        "metrics": [
            {
                "id": "00000000-0000-0000-0000-000000000000",
                "name": "rate that city is Boston and predicted class_a",
                "query": {
                    "select": [
                        {
                            "function": "rate",
                            "parameters": {
                                "property": "class_a",
                                "comparater": "gte",
                                "value": 0.5
                            }
                        }
                    ],
                    "filter": [
                        {
                            "property": "city",
                            "comparator": "eq",
                            "value": "Boston"
                        }
                    ]
                },
                "endpoint": "/api/v3/models/00000000-0000-0000-0000-000000000000/inferences/query",
                "is_default": False
            },
            {
                "id": "10000000-0000-0000-0000-000000000000",
                "name": "Inference Count",
                "query": {
                    "select": [
                        {
                            "function": "count"
                        }
                    ]
                },
                "endpoint": "/api/v3/models/10000000-0000-0000-0000-000000000000/inferences/query",
                "type": "model_output_metric",
                "is_default": True
            }
        ],
        "page": TestMetrics.current_page_num,
        "page_size": 2,
        "total_pages": 2,
        "total_count": 4
    }

    TestMetrics.current_page_num += 1

    return MockResponse(metric_list, 200)


def patch_metric_get_request_no_metrics_resp(*args, **kwargs):
    assert args[1] == f'None/models/4a57c553-e787-4307-88f1-3747ec9130f5/metrics'
    metrics = {"metrics":[],"page":1,"page_size":20}
    return MockResponse(metrics, 200)


def patch_metric_get_by_id_request(*args, **kwargs):
    assert args[1] == f'None/models/4a57c553-e787-4307-88f1-3747ec9130f5/metrics/{metric_id}'
    metric = {
        "id": metric_id,
        "name": "rate that city is Boston and predicted class_a",
        "query": {
            "select": [
                {
                    "function": "rate",
                    "parameters": {
                        "property": "class_a",
                        "comparater": "gte",
                        "value": 0.5
                    }
                }
            ],
            "filter": [
                {
                    "property": "city",
                    "comparator": "eq",
                    "value": "Boston"
                }
            ]
        },
        "endpoint": "/api/v3/models/00000000-0000-0000-0000-000000000000/inferences/query",
        "is_default": False
    }

    return MockResponse(metric, 200)


class TestMetrics:
    current_page_num = 1

    def test_create_metric(self, monkeypatch, mock_cred_env_vars, binary_classification_model: ArthurModel):
        query = {
            "select": [
                {
                    "function": "rate",
                    "parameters": {
                        "property": "class_a",
                        "comparater": "gte",
                        "value": 0.5
                    }
                }
            ],
            "filter": [
                {
                    "property": "city",
                    "comparator": "eq",
                    "value": "Boston"
                }
            ]
        }
        monkeypatch.setattr(BaseApiClient, "post", patch_metric_post_request)
        res = binary_classification_model.create_metric("Custom Metric A", query=query)
        assert res == metric_id

    def test_get_metrics_pagination(self, monkeypatch, mock_cred_env_vars, binary_classification_model: ArthurModel):
        monkeypatch.setattr(BaseApiClient, "get", patch_metric_get_request)
        res = binary_classification_model.get_metrics()
        assert len(res) == 4
        assert TestMetrics.current_page_num == 3

    def test_get_metrics_no_resp(self, monkeypatch, mock_cred_env_vars, binary_classification_model: ArthurModel):
        monkeypatch.setattr(BaseApiClient, "get", patch_metric_get_request_no_metrics_resp)
        res = binary_classification_model.get_metrics()
        assert len(res) == 0

    def test_get_metrics_error(self, monkeypatch, mock_cred_env_vars,
                               binary_classification_model: ArthurModel):
        monkeypatch.setattr(BaseApiClient, "get", patch_metric_get_request)
        with pytest.raises(ArthurUserError) as e:
            res = binary_classification_model.get_metrics(metric_type="invalid metric type")
        assert str(MetricType.list()) in str(e.value)

    def test_get_metric_by_id(self, monkeypatch, mock_cred_env_vars, binary_classification_model: ArthurModel):
        monkeypatch.setattr(BaseApiClient, "get", patch_metric_get_by_id_request)
        res = binary_classification_model.get_metrics(metric_id=metric_id)
        assert len(res) == 1
        assert res[0].id == metric_id



from typing import Optional, List

from arthurai.common.constants import InputType, OutputType, TextDelimiter
from arthurai.common.exceptions import MissingParameterError, UserValueError, arthur_excepted
from arthurai.core.models import ArthurModel
from arthurai.client.base import BaseApiClient
from arthurai.client.validation import validate_response_status

from http import HTTPStatus


class ArthurAI(object):
    """A client that interacts with Arthur's servers."""

    def __init__(self, config=None, verify_ssl=True, *args, **kwargs):
        if config:
            self.client = BaseApiClient(base_path='/api/v3', verify_ssl=verify_ssl, **config)
        else:
            self.client = BaseApiClient(base_path='/api/v3', verify_ssl=verify_ssl, *args, **kwargs)

    def model(self,
              partner_model_id: str,
              input_type: InputType,
              output_type: Optional[OutputType] = None,
              model_type: Optional[OutputType] = None,
              display_name: Optional[str] = None,
              description: Optional[str] = None,
              tags: Optional[List[str]] = None,
              classifier_threshold: Optional[float] = None,
              is_batch: bool = False,
              text_delimiter: Optional[TextDelimiter] = None,
              expected_throughput_gb_per_day: Optional[int] = None,
              pixel_height: Optional[int] = None,
              pixel_width: Optional[int] = None,) -> ArthurModel:
        """Create a new multistage model.

        :param: partner_model_id: The string external id of the model
        :param: input_type: the :py:class:`.InputType`
        :param: output_type: the :py:class:`.OutputType`
        :param: model_type: .. deprecated:: version 2.0.0 Use `output_type` instead.
        :param: display_name: Optional name to display on dashboard, will default to the external id
        :param: description: Optional description for the model
        :param: tags: A list of string tags to associate with the model
        :param: classifier_threshold: For binary classification models this is the threshold to determine a positive class, defaults to 0.5
        :param: is_batch boolean value which signifies whether the model sends inferences by batch or streaming
        :param: text_delimiter TextDelimiter used in NLP models to split documents into tokens for explanations
        :param: expected_throughput_gb_per_day: Expected amount of throughput. Used to provision resources
        :param: pixel_height: Image height in pixels. Needed for CV models which require images to be one size
        :param: pixel_width: Image width in pixels. Needed for CV models which require images to be one size

        :return: An :py:class:`~arthurai.client.apiv3.model.ArthurModel`
        """
        if output_type is None and model_type is None:
            raise MissingParameterError("Either 'output_type' or 'model_type' parameter must be specified")
        output_type = output_type if output_type is not None else model_type

        return ArthurModel(
            client=self.client,
            partner_model_id=partner_model_id,
            input_type=input_type,
            output_type=output_type,
            display_name=display_name,
            description=description,
            tags=tags,
            classifier_threshold=classifier_threshold,
            is_batch=is_batch,
            text_delimiter=text_delimiter,
            expected_throughput_gb_per_day=expected_throughput_gb_per_day,
            pixel_height=pixel_height,
            pixel_width=pixel_width,
        )

    @arthur_excepted("failed to retrieve model")
    def get_model(self, identifier: str, id_type: str = 'id') -> ArthurModel:
        """Retrieve an existing model by id

        :param: identifier: Id to get the model by
        :param: id_type: Type of id the identifier is, possible options are ['id', 'partner_model_id']

        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if id_type in ['partner_model_id', 'id']:
            url = self.client.base_path+f'/models/{identifier}?id_type={id_type}&expand=attributes'
        else:
            raise UserValueError(f"Invalid id_type: {id_type}, must be one of the following ['id', 'partner_model_id]")

        resp = self.client.get(url, return_raw_response=True)
        validate_response_status(resp, expected_status_code=HTTPStatus.OK)

        # accuracy data is only used by the UI not needed in the SDK
        model_data = resp.json()
        if "accuracy_enabled" in model_data:
            del model_data["accuracy_enabled"]

        model = ArthurModel.from_dict(model_data)
        model._client = self.client
        return model



import re
import json
from typing import Union, Dict, List
from requests import Response
from http import HTTPStatus

import pytest
import responses
from unittest import mock

from tests import MockResponse

from arthurai import ArthurAI
from arthurai.common.constants import InputType, OutputType
from arthurai.client.validation import validate_multistatus_response_and_get_failures, validate_response_status
from arthurai.common.exceptions import ResponseServerError, ResponseClientError, ResponseRedirectError, \
    InternalValueError, UserValueError
from tests.base_test import BaseTest


class TestClient(BaseTest):

    def _mock_get_user(self):
        self.mockPost(f"/api/v3/users/me", request_type=responses.GET, status=HTTPStatus.OK,
                      dict_to_return={"organization_id": "65624ee6-9a73-4fe3-a7ff-8612f7d11c21", "role": "Model Owner"})

    def _mock_get_user_unauthorized(self):
        self.mockPost(f"/api/v3/users/me", request_type=responses.GET, status=HTTPStatus.UNAUTHORIZED,
                      dict_to_return={"bad": "user"})

    @pytest.mark.usefixtures('mock_cred_env_vars')
    def test_client_init_env_vars(self):
        client = ArthurAI(offline=True)
        assert client.client.url == 'http://mock'
        assert client.client.access_key == 'access_key'

        client = ArthurAI(url="this is a url", access_key="this is an access key", offline=True)
        assert client.client.url == "https://this is a url"
        assert client.client.access_key == "this is an access key"

        client = ArthurAI(url="http://this is a url", access_key="this is an access key", offline=True)
        assert client.client.url == "http://this is a url"
        assert client.client.access_key == "this is an access key"

        client = ArthurAI(url="http://this is a url/apath", access_key="this is an access key", offline=True)
        assert client.client.url == "http://this is a url"
        assert client.client.access_key == "this is an access key"

        client = ArthurAI(url="http://this is a url?food=good", access_key="this is an access key", offline=True)
        assert client.client.url == "http://this is a url"
        assert client.client.access_key == "this is an access key"

        client = ArthurAI(url="fatfinger://this is a url", access_key="this is an access key", offline=True)
        assert client.client.url == "https://this is a url"
        assert client.client.access_key == "this is an access key"

        with pytest.raises(ValueError):
            ArthurAI(url="https:", access_key="this is an access key", offline=True)

    @responses.activate
    def test_client_init(self):
        self._mock_get_user()

        ArthurAI(access_key=self.access_key, url=self._base_url)
        self.assertEqual(len(responses.calls), 1)
        expectedResponse = {'organization_id': '65624ee6-9a73-4fe3-a7ff-8612f7d11c21', 'role': 'Model Owner'}
        self.assertEqual(expectedResponse, responses.calls[0].response.json())

    @responses.activate
    def test_client_init_bad_host(self):
        with pytest.raises(UserValueError, match=".*please ensure the URL is correct.*"):
            ArthurAI(access_key=self.access_key, url=self._base_url)
        self.assertEqual(len(responses.calls), 1)

    @responses.activate
    def test_client_init_bad_access_key(self):
        self._mock_get_user_unauthorized()
        with pytest.raises(UserValueError, match=".*please ensure your access key is correct.*"):
            ArthurAI(access_key=self.access_key, url=self._base_url)
        self.assertEqual(len(responses.calls), 1)

    @responses.activate
    def test_client_init_offline(self):
        ArthurAI(access_key=self.access_key, url=self._base_url, offline=True)
        self.assertEqual(len(responses.calls), 0)

    @responses.activate
    def test_client_init_offlines(self):
        self.mockPost(f"/api/v3/models/1234", request_type=responses.GET, status=200,
                      dict_to_return={"organization_id": "65624ee6-9a73-4fe3-a7ff-8612f7d11c21", "role": "Model Owner"})
        ArthurAI(access_key=self.access_key, url=self._base_url, offline=True)
        self.assertEqual(len(responses.calls), 0)

    @responses.activate
    def test_client_user_agent_header(self):
        model_data = {
            "partner_model_id": "",
            "input_type": InputType.Tabular,
            "output_type": OutputType.Regression,
            "display_name": "",
            "description": "",
            "attributes": [],
        }
        self.mockPost("/api/v3/models/1234", request_type=responses.GET, status=200,
                      dict_to_return=model_data)

        self._mock_get_user()

        client = ArthurAI(access_key=self.access_key, url=self._base_url)
        client.get_model('1234')

        self.assertEqual(len(responses.calls), 2)
        if "models" in responses.calls[1].request.url:
            header = responses.calls[1].request.headers
        else:
            header = responses.calls[0].request.headers
        self.assertTrue('User-Agent' in header)
        print(header['User-Agent'])
        assert re.findall(r"^arthur-sdk\/\d.\d+.\d+ \(system=\w*, org=65624ee6-9a73-4fe3-a7ff-8612f7d11c21\)$",
                          header['User-Agent'])


class TestGetMultiStatusResponseFailures:
    # fixtures
    SUCCESS_RESULTS = [{"message": "ok", "status": HTTPStatus.OK}, {"message": "created", "status": HTTPStatus.CREATED}]
    USER_FAIL_RESULTS = [{"message": "not found", "status": HTTPStatus.NOT_FOUND}]
    INTERNAL_FAIL_RESULTS = [{"message": "failure", "status": HTTPStatus.INTERNAL_SERVER_ERROR},
                             {"message": "choose", "status": HTTPStatus.MULTIPLE_CHOICES}]
    SUCCESS_BODY = {"counts": {"success": len(SUCCESS_RESULTS), "failure": 0, "total": len(SUCCESS_RESULTS)},
                    "results": SUCCESS_RESULTS}
    FAILURES_BODY = {"counts": {"success": len(SUCCESS_RESULTS),
                                "failure": len(INTERNAL_FAIL_RESULTS) + len(USER_FAIL_RESULTS),
                                "total": len(SUCCESS_RESULTS) + len(INTERNAL_FAIL_RESULTS) + len(USER_FAIL_RESULTS)},
                     "results": SUCCESS_RESULTS + USER_FAIL_RESULTS + INTERNAL_FAIL_RESULTS}
    MISMATCHED_COUNTS_BODY = {"counts": {"success": len(SUCCESS_RESULTS), "failure": 0,
                                         "total": len(SUCCESS_RESULTS)},
                              "results": SUCCESS_RESULTS + USER_FAIL_RESULTS + INTERNAL_FAIL_RESULTS}

    def test_success(self):
        response = MockResponse(self.SUCCESS_BODY, HTTPStatus.MULTI_STATUS)
        actual_user_failures, actual_internal_failures = validate_multistatus_response_and_get_failures(response)
        assert actual_user_failures == []
        assert actual_internal_failures == []

    def test_failures(self):
        response = MockResponse(self.FAILURES_BODY, HTTPStatus.MULTI_STATUS)
        actual_user_failures, actual_internal_failures = validate_multistatus_response_and_get_failures(response)
        assert actual_user_failures == self.USER_FAIL_RESULTS
        assert actual_internal_failures == self.INTERNAL_FAIL_RESULTS

    def test_bad_status(self):
        response = MockResponse(self.SUCCESS_BODY, HTTPStatus.OK)
        with pytest.raises(ValueError):
            validate_multistatus_response_and_get_failures(response)

    def test_mismatched_counts(self):
        response = MockResponse(self.MISMATCHED_COUNTS_BODY, HTTPStatus.MULTI_STATUS)
        with pytest.raises(ValueError):
            validate_multistatus_response_and_get_failures(response)

# TODO: add testing around allowed exceptions and what gets raised first (e.g. expected 207 but 404 shouldn't raise InternalValueError)
class TestValidateResponse:
    # fixtures
    JSON_BODY = {"status": "somestatus"}
    BYTES_BODY = b"bodybytes"

    def test_server_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.INTERNAL_SERVER_ERROR)
        with pytest.raises(ResponseServerError):
            validate_response_status(response)

    def test_client_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.NOT_FOUND)
        with pytest.raises(ResponseClientError):
            validate_response_status(response)

    def test_expected_404_no_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.NOT_FOUND)
        validate_response_status(response, expected_status_code=HTTPStatus.NOT_FOUND)

    def test_redirect_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.MOVED_PERMANENTLY)
        with pytest.raises(ResponseRedirectError):
            validate_response_status(response)

    def test_redirect_no_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.MOVED_PERMANENTLY)
        validate_response_status(response, allow_redirects=True)

    def test_ok_no_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.OK)
        validate_response_status(response)

    def test_unexpected_ok_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.OK)
        with pytest.raises(InternalValueError):
            validate_response_status(response, expected_status_code=HTTPStatus.CREATED)

    def test_server_error(self):
        code = HTTPStatus.INTERNAL_SERVER_ERROR
        with pytest.raises(ResponseServerError):
            validate_response_status(code)

    def test_client_error(self):
        code = HTTPStatus.NOT_FOUND
        with pytest.raises(ResponseClientError):
            validate_response_status(code)

    def test_expected_404_no_error(self):
        code = HTTPStatus.NOT_FOUND
        validate_response_status(code, expected_status_code=HTTPStatus.NOT_FOUND)

    def test_redirect_error(self):
        code = HTTPStatus.MOVED_PERMANENTLY
        with pytest.raises(ResponseRedirectError):
            validate_response_status(code)

    def test_redirect_no_error(self):
        code = HTTPStatus.MOVED_PERMANENTLY
        validate_response_status(code, allow_redirects=True)

    def test_ok_no_error(self):
        code = HTTPStatus.OK
        validate_response_status(code)

    def test_unexpected_ok_error(self):
        code = HTTPStatus.OK
        with pytest.raises(InternalValueError):
            validate_response_status(code, expected_status_code=HTTPStatus.CREATED)



import pandas as pd 
import matplotlib.pyplot as plt
import seaborn as sns

from arthurai.common.exceptions import arthur_excepted
from arthurai.core.viz import style, utils
import warnings
warnings.filterwarnings('ignore')


class DataVisualizer(object):
    def __init__(self, arthur_model):
        self.model = arthur_model

    @arthur_excepted("failed to generate timeline")
    def timeline(self, attribute_name):
        """Generates a visualization of the distribution of an attribute over time. 
        For categorical attributes, a stacked area chart over time.
        For continuous attributes, a joyplot showing probability densities over time.
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if self.model.get_attribute(attribute_name).categorical:
            self._timeline_categorical(self.model, attribute_name)
        else:
            self._timeline_continuous(self.model, attribute_name)

    @arthur_excepted("failed to generate metric series")
    def metric_series(self, metric_names, time_resolution="day"):
        """Generates a line series visualization for selected metrics.
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        fig = plt.figure(figsize=(20,4))
        ax = fig.add_subplot(1, 1, 1)
        predicted_property, ground_truth_property = utils.get_pred_and_gt_attrs(self.model)
        pal = style.categorical_palette

        for index, metric_name in enumerate(metric_names):
            if self.model.is_batch:
                time_resolution = "batch_id"
                query = _metric_series_batch_query(metric_name, ground_truth_property, 
                    predicted_property, self.model.classifier_threshold)
            else:
                query = _metric_series_streaming_query(metric_name, time_resolution, 
                    ground_truth_property, predicted_property, self.model.classifier_threshold)

            response = self.model.query(query)
            df = pd.DataFrame(response)[::-1]

            plt.plot(df[time_resolution], df[metric_name], lw=12, color=pal[index], label=metric_name)
            plt.plot(df[time_resolution], df[metric_name], lw=2, color=pal[index])
            plt.xticks([])
            ax.spines['right'].set_visible(False)
            ax.spines['top'].set_visible(False)
            plt.legend(loc="upper left", facecolor="white")

    @arthur_excepted("failed to generate drift series")
    def drift_series(self, attribute_names, drift_metric="PSI", time_resolution="day"):
        """Generates a line series visualization of data drift metrics for selected attributes.
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        fig = plt.figure(figsize=(20,4))
        ax = fig.add_subplot(1, 1, 1)

        if self.model.is_batch:
            time_resolution = "batch_id"
        timestamp_query = _batch_timestamp_query()
        drift_query = {
            "properties": attribute_names,
            "num_bins": 20,
            "rollup": time_resolution,
            "base": {
            "source": "reference"
            },
            "target": {
            "source": "inference"
            },
            "metric": drift_metric
        }
        if self.model.is_batch:
            drift_df = pd.DataFrame(self.model.query(drift_query, query_type="drift"))
            timestamp_df = pd.DataFrame(self.model.query(timestamp_query))
            df = drift_df.rename(columns={"rollup":time_resolution}
                ).join(timestamp_df.set_index(time_resolution), on=time_resolution
                ).sort_values(by="timestamp"
                ).set_index(time_resolution)
        else:
            df = pd.DataFrame(self.model.query(drift_query, query_type="drift")
                    ).sort_values(by="rollup").set_index("rollup")

        df[attribute_names].plot(ax=ax, color=style.categorical_palette, lw=8)
        plt.ylabel(drift_metric)
        plt.xticks([])
        plt.xlabel("")
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        plt.legend(loc="upper left", facecolor="white")

    def _timeline_categorical(self, model, attribute_name, time_resolution="day"):
        """
        """
        if self.model.is_batch:
            time_resolution="timestamp"
            query = _timeline_categorical_batch_query(attribute_name, time_resolution)
        else:
            query = _timeline_categorical_streaming_query(attribute_name, time_resolution)
        response = self.model.query(query)
        df = pd.DataFrame(response
            ).sort_values(by=time_resolution
            ).pivot(index=time_resolution,columns=attribute_name, values="count"
            ).fillna(0.0)
        
        sns.set(style="white")
        fig = plt.figure(figsize=(20,4))
        ax = fig.add_subplot(1, 1, 1)
        plt.stackplot(df.index.values, df.values.T.tolist(), 
            labels=df.columns.values, colors=style.categorical_palette)
        plt.xticks([])
        plt.ylabel("Count")
        plt.legend(loc='upper right')
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        plt.show()



    def _timeline_continuous(self, model, attribute_name, time_resolution="day"):
        """
        """
        if self.model.is_batch:
            time_resolution="timestamp"
            query = _timeline_continuous_batch_query(attribute_name, time_resolution)
        else:
            query = _timeline_continuous_streaming_query(attribute_name, time_resolution)
        response = model.query(query)
        
        dfs= []
        for group in response:
            temp_df= pd.DataFrame(group["distribution"])
            temp_df[time_resolution] = group[time_resolution]
            temp_df["count"] = utils.savgol_filter(temp_df["count"])
            dfs.append( temp_df)
        df = pd.concat(dfs).sort_values([time_resolution, "lower"])
        
        global_y_min = df["lower"].min()
        global_y_max = df["lower"].max()
        global_x_min = df["count"].min()
        global_x_max = df["count"].max()
        
        sns.set(style="white", rc={"axes.facecolor": (0, 0, 0, 0)})
        # Initialize the FacetGrid object
        pal = style.continuous_palette
        g = sns.FacetGrid(df, col=time_resolution, hue=time_resolution, aspect=.2, height=8, palette=pal, subplot_kws={"xlim":(global_x_max, global_x_min), "ylim":(global_y_min, global_y_max)})

        # Draw the densities in a few steps
        g.map(plt.plot, "count", "lower",clip_on=False,  color="white", lw=4.5,  alpha=0.6)
        g.map(plt.fill_between, "count","lower", clip_on=False, lw=2)

        # Set the subplots to overlap
        g.fig.subplots_adjust(wspace=-.55, left = 0.125)
        g.set_titles("")
        g.set(yticks=[])
        g.set(xticks=[])
        g.set(ylabel="")
        g.set(xlabel="")
        g.despine(bottom=True, left=True)


def _metric_series_streaming_query(metric_name, time_resolution, ground_truth_property, predicted_property, classifier_threshold):
    """Generates query for fetching metrics for streaming model.
    """
    query = {
                "select": [
                    {
                        "function": metric_name,
                        "alias": metric_name,
                        "parameters": {
                            "ground_truth_property": ground_truth_property,
                            "predicted_property": predicted_property,
                            "threshold": classifier_threshold
                        }

                    },

                    {
                        "function": "roundTimestamp",
                        "alias": time_resolution,
                        "parameters": {
                            "property": "inference_timestamp",
                            "time_interval": time_resolution
                        }
                    }
                ],
                "group_by": [
                    {
                        "alias": time_resolution
                    }
                ],
                "order_by": [
                    {
                        "alias": time_resolution,
                        "direction": "desc"
                    }
                ]
            }
    return query

def _metric_series_batch_query(metric_name, ground_truth_property, predicted_property, classifier_threshold):
    """Generates query for fetching metrics for batch model.
    """
    query = {
                "select": [
                    {"function": metric_name,
                        "alias": metric_name,
                        "parameters": {
                            "ground_truth_property": ground_truth_property,
                            "predicted_property": predicted_property,
                            "threshold": classifier_threshold
                        }

                    },
                    {
                        "function":"max", 
                        "parameters": {"property":"inference_timestamp"}, 
                        "alias":"timestamp"
                    },
                    {"property":"batch_id"}
                ],
                "group_by": [
                    {
                        "property":"batch_id"
                    }
                ],
                "order_by": [
                    {
                        "alias":"timestamp",
                        "direction": "desc"
                    }
                ]
            }
    return query

def _timeline_continuous_streaming_query(attribute_name, time_resolution):
    """Generates a query for continuous attribute in a streaming model.
    """
    return {
        "select":[
            {
                "function":"distribution", 
                "alias":"distribution", 
                "parameters":{"property": attribute_name, "num_bins":50}
            },
            {
                "function": "roundTimestamp",
                "alias": time_resolution,
                "parameters": {
                    "property": "inference_timestamp",
                    "time_interval": time_resolution
                    }
            }
        ],
        "group_by": [{"alias":time_resolution} ],
        "order_by": [{"alias":time_resolution, "direction": "desc"} ]
    }

def _timeline_continuous_batch_query(attribute_name, time_resolution):
    """Generates a query for continuous attribute in a batch model.
    """
    return {
        "select":[
            {
                "function":"distribution", 
                "alias":"distribution", 
                "parameters":{"property": attribute_name, "num_bins":50}
            },
            {
                "function":"max", 
                "alias":"timestamp",
                "parameters": {"property":"inference_timestamp"}
            },
            {"property": "batch_id"}
        ],
        "group_by": [
            {
                "property":"batch_id"
            }
        ],
        "order_by": [
            {
                "alias":"timestamp",
                "direction": "desc"
            }
        ]
    }

def _timeline_categorical_streaming_query(attribute_name, time_resolution):
    """Generates a query for categorical attribute in a streaming model.
    """
    return {
        "select": [
            {
                "property": attribute_name
            },
            {
                "function": "count"
            },
            {
                "function": "roundTimestamp",
                "alias": time_resolution,
                "parameters": {
                    "property": "inference_timestamp",
                    "time_interval": time_resolution
                }
            }
        ],
        "group_by": [
            {
                "property": attribute_name
            },
            {
                "alias": time_resolution
            }
        ],
        "order_by": [
            {
                "property": attribute_name
            },
            {
                "alias": time_resolution,
                "direction": "desc"
            }
        ]
    }

def _timeline_categorical_batch_query(attribute_name, time_resolution):
    """Generates a query for categorical attribute in a batch model.
    """
    return {
        "select": [
            {
                "property": attribute_name
            },
            {
                "function": "count"
            },
            {
                "property":"batch_id"
            },
            {"function":"max", "parameters": {"property":"inference_timestamp"}, "alias":"timestamp"}
        ],
        "group_by": [
            {
                "property": attribute_name
            },
            {
                "property":"batch_id"
            }
        ],
        "order_by": [
            {
                "property": attribute_name
            },
            {
                "alias":"timestamp",
                "direction": "desc"
            }
        ]
    }

def _batch_timestamp_query():
    return {
        "select":[
            {
                "function":"max", 
                "alias":"timestamp",
                "parameters": {"property":"inference_timestamp"}
            },
            {"property": "batch_id"}
        ],
        "group_by": [
            {
                "property":"batch_id"
            }
        ],
        "order_by": [
            {
                "alias":"timestamp",
                "direction": "desc"
            }
        ]
    }


import pytest

from arthurai import ArthurModel, ArthurAttribute
from arthurai.common.constants import InputType, OutputType, ValueType, Stage


TABULAR_MODEL_ID = "tabby-mc-tabface"
INT_INPUT = "int_input"
INT_INPUT_ATTR = ArthurAttribute(name=INT_INPUT, value_type=ValueType.Integer, stage=Stage.ModelPipelineInput,
                                 categorical=True)
FLOAT_INPUT = "float_input"
FLOAT_INPUT_ATTR = ArthurAttribute(name=FLOAT_INPUT, value_type=ValueType.Float, stage=Stage.ModelPipelineInput)
PRED = "pred"
PRED_ATTR = ArthurAttribute(name=PRED, value_type=ValueType.Float, stage=Stage.PredictedValue)
GROUND_TRUTH = "ground_truth"
GROUND_TRUTH_ATTR = ArthurAttribute(name=GROUND_TRUTH, value_type=ValueType.Float, stage=Stage.GroundTruth)


@pytest.fixture
def tabular_model(mock_client):
    model_data = {
        "partner_model_id": "",
        "input_type": InputType.Tabular,
        "output_type": OutputType.Regression,
        "display_name": "",
        "description": "",
        "attributes": [INT_INPUT_ATTR, FLOAT_INPUT_ATTR, PRED_ATTR, GROUND_TRUTH_ATTR],
    }
    model = ArthurModel(client=mock_client.client, **model_data)
    model.id = TABULAR_MODEL_ID
    return model



class ListableStrEnum:
    """
    This class operates similar to Enum but passes mypy type checking.
    The members are accessed directly rather than by `.name` or `.value`.
    """

    @classmethod
    def list(cls):
        """
        Lists all attributes in alphabetical order
        """
        members = [getattr(cls, attr) for attr in dir(cls)
                   if not callable(getattr(cls, attr)) and not attr.startswith("__")]
        return sorted(members)


class InputType(ListableStrEnum):
    Tabular = "TABULAR"
    Image = "IMAGE"
    NLP = "NLP"


class OutputType(ListableStrEnum):
    Regression = "REGRESSION"
    Multiclass = "MULTICLASS"
    Multilabel = "MULTILABEL"
    ObjectDetection = "OBJECT_DETECTION"


class ValueType(ListableStrEnum):
    String = "STRING"
    Integer = "INTEGER"
    Float = "FLOAT"
    Image = "IMAGE"
    Boolean = "BOOLEAN"
    Timestamp = "TIMESTAMP"
    Unstructured_Text = "UNSTRUCTURED_TEXT" # don't remove old one, backward compatibility
    BoundingBox = "BOUNDING_BOX"


class Stage(ListableStrEnum):
    ModelPipelineInput = "PIPELINE_INPUT"
    PredictFunctionInput = "PREDICT_FUNCTION_INPUT"
    PredictedValue = "PREDICTED_VALUE"
    NonInputData = "NON_INPUT_DATA"
    GroundTruth = "GROUND_TRUTH"


class TextDelimiter(ListableStrEnum):
    """ This class contains patterns that can be used
    as text_delimiter for NLP models.

    Definitions:

    NOT_WORD: `'\W+'` Splits on any character that is not a word. Ex:
              "this  is,aaaa,,,,,test!" = ["this", "is", "aaaa", "test", ""]

    WHITESPACE: `'\s+` Splits on whitespace. Ex:
              "this  is,a test! " = ["this", "is,a", "test!", ""]

    COMMA: `","` Splits on a single comma. Ex:
              "this ,is,,a,test" = ["this ", "is", "", "a", "test"]

    COMMA_PLUS: `",+"` Splits on one or more commas. Ex:
              "this ,is,,a,test" = ["this ", "is", "a", "test"]

    PIPE: `"\|"` Splits on a single pipe. Ex:
              "this |is||a|test" = ["this ", "is", "", "a", "test"]

    PIPE_PLUS: `"\|+"` Splits on one or more pipes. Ex:
              "this |is||a|test" = ["this ", "is", "a", "test"]
    """
    NOT_WORD = "\W+"
    WHITESPACE = "\s+"
    COMMA = ","
    COMMA_PLUS = ",+"
    PIPE = "\|"
    PIPE_PLUS = "\|+"


class Enrichment(ListableStrEnum):
    """This class contains constants for the names of enrichments"""
    AnomalyDetection = "anomaly_detection"
    Explainability = "explainability"


class ImageResponseType(ListableStrEnum):
    """Valid image response types supported by Arthur API"""
    RawImage = 'raw_image'
    ResizedImage = 'resized_image'
    Thumbnail = 'thumbnail'
    LimeExplanation = 'lime_explanation'


class ImageContentType(ListableStrEnum):
    """Valid image content types supported by Arthur API"""
    Png = 'image/png'
    Jpeg = 'image/jpeg'
    Gif = 'image/gif'
    Tiff = 'image/tiff'


IMAGE_FILE_EXTENSION_MAP = {
    ImageContentType.Png: '.png',
    ImageContentType.Jpeg: '.jpg',
    ImageContentType.Gif: '.gif',
    ImageContentType.Tiff: '.tff'
}



from typing import Any, Dict

from setuptools import setup, find_packages

version: Dict[str, Any] = {}
with open("./arthurai/version.py") as fp:
    exec(fp.read(), version)

setup(
    name='arthurai',
    version=version['__version__'],
    description='Python SDK for ArthurAI',
    url='https://arthur.ai',
    author='ArthurAI',
    author_email='info@arthur.ai',
    license='MIT',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

    keywords='api arthur ArthurAI sdk',

    packages=find_packages(),
    include_package_data=True,

    install_requires=[
        'dataclasses==0.7;python_version<"3.7"',
        'dataclasses-json==0.4.5',
        'azure-storage-file-datalake==12.0.2',
        'dill>=0.3.1.1,<=0.3.3',
        'docker>=4.2.0,<4.4.0',
        'requests>=2.13.0,<=2.25.0',
        'numpy>=1.16.0,<=1.20.1',
        'pandas>=1.0.0,<=1.2.1',
        'lime~=0.2.0.0',
        'cloudpickle>=1.2.0,<=1.6.0',
        'pyarrow>=0.16.0,<=3.0.0',
        'python-dateutil>=2.7.0,<2.9',
        'shap==0.35.0',
        'shortuuid==1.0.1',
        'seaborn==0.11.1'
    ]
)



from collections import namedtuple
from typing import Any, Type, Optional, Union, List, NamedTuple

import os
from unittest.mock import patch

import pandas as pd
import pytest
import responses

from arthurai import ArthurModel
from arthurai.common.exceptions import UserValueError
from arthurai.core.inferences import add_inference_metadata_to_dataframe
from tests.helpers import mock_post, MockShortUUID, MockDatetime, mock_patch
from tests.fixtures.mocks import mock_client
from tests.fixtures.models import tabular_model, INT_INPUT, FLOAT_INPUT, PRED, GROUND_TRUTH, TABULAR_MODEL_ID


class TestBulkInferences:

    INT_INPUT_VALUES = [1, 2, 3, 4]
    FLOAT_INPUT_VALUES = [1.1, 2.2, 3.3, 4.4]
    PRED_VALUES = [0.6, 0.4, 0.7, 0.3]
    GROUND_TRUTH_VALUES = [1, 1, 1, 0]
    TIMESTAMP_VALUE = 'ts1'
    INFERENCE_TIMESTAMP_VALUES = [TIMESTAMP_VALUE] * 4
    GROUND_TRUTH_TIMESTAMP_VALUES = [TIMESTAMP_VALUE] * 4
    PARTNER_INFERENCE_ID_VALUES = ['inf1', 'inf2', 'inf3', 'inf4']

    INFERENCE_DATA = {
        INT_INPUT: INT_INPUT_VALUES,
        FLOAT_INPUT: FLOAT_INPUT_VALUES,
        PRED: PRED_VALUES
    }

    INFERENCE_ONLY_DF = pd.DataFrame({**INFERENCE_DATA, 'inference_timestamp': INFERENCE_TIMESTAMP_VALUES,
                                      'partner_inference_id': PARTNER_INFERENCE_ID_VALUES})
    INFERENCE_AND_GT_DF = pd.DataFrame({**INFERENCE_DATA, 'inference_timestamp': INFERENCE_TIMESTAMP_VALUES,
                                        'partner_inference_id': PARTNER_INFERENCE_ID_VALUES,
                                        GROUND_TRUTH: GROUND_TRUTH_VALUES,
                                        'ground_truth_timestamp': GROUND_TRUTH_TIMESTAMP_VALUES})
    GT_ONLY_DF = pd.DataFrame({'partner_inference_id': PARTNER_INFERENCE_ID_VALUES, GROUND_TRUTH: GROUND_TRUTH_VALUES,
                               'ground_truth_timestamp': GROUND_TRUTH_TIMESTAMP_VALUES})

    def _mock_sending_reference_set(self, model_id):
        server_response = {
            "counts": {
                "success": 0,
                "failure": 1,
                "total": 1
            },
            "failures": [
                {
                    "message": "missing field",
                    "status": 400
                }
            ]
        }

        mock_post(f"/api/v3/models/{model_id}/reference_data", server_response, status=207)
        mock_patch(f"/api/v3/models/{model_id}/reference_data", status=200,
                   dict_to_return={"message": "success"})

    def _mock_batch_upload(self, model_id, batch_id):
        server_response = {
            "counts": {
                "success": 100,
                "failure": 0,
                "total": 100
            },
            "failures": []
        }

        mock_post(f"/api/v3/models/{model_id}/inferences/file", server_response, status=207)
        mock_patch(f"/api/v3/models/{model_id}/batches/{batch_id}", status=200,
                   dict_to_return={"message": "success"})

    def _mock_failed_batch_upload(self, model_id, batch_id):
        mock_post(f"/api/v3/models/{model_id}/inferences/file", {}, status=400)

    def _mock_failed_batch_close(self, model_id, batch_id):
        server_response = {
            "counts": {
                "success": 100,
                "failure": 0,
                "total": 100
            },
            "failures": []
        }

        mock_post(f"/api/v3/models/{model_id}/inferences/file", server_response, status=207)
        mock_patch(f"/api/v3/models/{model_id}/batches/{batch_id}", status=404,
                   dict_to_return={"message": "batch not found"})

    @responses.activate
    def test_send_set_reference_data(self, tabular_model):
        self._mock_sending_reference_set(TABULAR_MODEL_ID)
        resp, resp_close = tabular_model.set_reference_data(data=pd.DataFrame(
            {**self.INFERENCE_DATA, GROUND_TRUTH: self.GROUND_TRUTH_VALUES}))
        assert resp["counts"]["failure"] == 1
        assert resp_close is None

    @responses.activate
    def test_send_bulk_inferences(self, tabular_model):
        self._mock_batch_upload(TABULAR_MODEL_ID, 'batch_1')
        resp, resp_close = tabular_model.send_bulk_inferences(
            data=pd.DataFrame({
                INT_INPUT: self.INT_INPUT_VALUES,
                FLOAT_INPUT: self.FLOAT_INPUT_VALUES,
                PRED: self.PRED_VALUES,
                "inference_timestamp": self.INFERENCE_TIMESTAMP_VALUES,
                "partner_inference_id": self.PARTNER_INFERENCE_ID_VALUES,
            }),
            batch_id='batch_1')
        self._assert_result(resp)
        assert resp_close == {"dataset_close_result": "success"}

    @responses.activate
    def test_send_bulk_inferences_auto_retry(self, tabular_model):
        self._mock_failed_batch_upload(TABULAR_MODEL_ID, "batch_1")
        resp, resp_close = tabular_model.send_bulk_inferences(
            data=pd.DataFrame({
                INT_INPUT: self.INT_INPUT_VALUES,
                FLOAT_INPUT: self.FLOAT_INPUT_VALUES,
                PRED: self.PRED_VALUES,
                "inference_timestamp": self.INFERENCE_TIMESTAMP_VALUES,
                "partner_inference_id": self.PARTNER_INFERENCE_ID_VALUES,
            }),
            batch_id='batch_1')
        assert len(responses.calls) == 4
        resp_400 = 0
        for http_call in responses.calls:
            if http_call.response.status_code == 400:
                resp_400 += 1
        assert resp_400 == 4

    @responses.activate
    def test_send_bulk_close_auto_retry(self, tabular_model):
        self._mock_failed_batch_close(TABULAR_MODEL_ID, "batch_1")
        resp, resp_close = tabular_model.send_bulk_inferences(
            data=pd.DataFrame({
                INT_INPUT: self.INT_INPUT_VALUES,
                FLOAT_INPUT: self.FLOAT_INPUT_VALUES,
                PRED: self.PRED_VALUES,
                "inference_timestamp": self.INFERENCE_TIMESTAMP_VALUES,
                "partner_inference_id": self.PARTNER_INFERENCE_ID_VALUES,
            }),
            batch_id='batch_1')
        assert len(responses.calls) == 5
        resp_404 = 0
        for http_call in responses.calls:
            if http_call.response.status_code == 404:
                resp_404 += 1
        assert resp_404 == 4
        assert resp_close == {'dataset_close_result': {"message": "batch not found"}}

    @responses.activate
    def test_send_bulk_ground_truth(self, tabular_model):
        self._mock_batch_upload(TABULAR_MODEL_ID, 'batch_1')
        resp = tabular_model.send_batch_ground_truths(
            data=pd.DataFrame({GROUND_TRUTH: self.GROUND_TRUTH_VALUES,
                               "partner_inference_id": self.PARTNER_INFERENCE_ID_VALUES,
                               "ground_truth_timestamp": self.GROUND_TRUTH_TIMESTAMP_VALUES}))
        self._assert_result(resp)

    @responses.activate
    def test_send_parquet_batch(self, tabular_model):
        self._mock_batch_upload(TABULAR_MODEL_ID, 'batch_1')
        file_loc = os.path.join(os.path.dirname(__file__), 'data', 'inference_batch_data')
        resp, close_resp = tabular_model.send_bulk_inferences(batch_id="batch_1", directory_path=file_loc)
        self._assert_result(resp)
        assert close_resp == {"dataset_close_result": "success"}

    # these cases are used  by two tests below: primarily for testing of the add_inference_metadata_to_dataframe()
    #  method but also include a broader test of send_bulk_inferences() to ensure no exceptions occur along the full
    #  processing chain
    class BulkInferencesCase(NamedTuple):
        input_dataframe: pd.DataFrame
        partner_inf_ids: Optional[List[str]]
        expected_result: Union[pd.DataFrame, Type[BaseException]]
        ignore_join_errors: Optional[bool] = None

    bulk_inferences_metadata_cases = [
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame({**INFERENCE_DATA, GROUND_TRUTH: GROUND_TRUTH_VALUES}),
            partner_inf_ids=PARTNER_INFERENCE_ID_VALUES, expected_result=INFERENCE_AND_GT_DF),
            id="inf-and-gt-data-no-metadata"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame({**INFERENCE_DATA, GROUND_TRUTH: GROUND_TRUTH_VALUES,
                                          'inference_timestamp': INFERENCE_TIMESTAMP_VALUES}),
            partner_inf_ids=PARTNER_INFERENCE_ID_VALUES, expected_result=INFERENCE_AND_GT_DF),
            id="inf-and-gt-data-inf-ts-only"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame({**INFERENCE_DATA, GROUND_TRUTH: GROUND_TRUTH_VALUES,
                                          'ground_truth_timestamp': GROUND_TRUTH_TIMESTAMP_VALUES}),
            partner_inf_ids=PARTNER_INFERENCE_ID_VALUES, expected_result=INFERENCE_AND_GT_DF),
            id="inf-and-gt-data-gt-ts-only"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame({**INFERENCE_DATA, GROUND_TRUTH: GROUND_TRUTH_VALUES,
                                          'partner_inference_id': PARTNER_INFERENCE_ID_VALUES}),
            partner_inf_ids=None, expected_result=INFERENCE_AND_GT_DF), id="inf-and-gt-data-gt-partner-ids-only"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame({GROUND_TRUTH: GROUND_TRUTH_VALUES,
                                          'partner_inference_id': PARTNER_INFERENCE_ID_VALUES}),
            partner_inf_ids=None, expected_result=GT_ONLY_DF), id="gt-data-only-no-ts"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame({GROUND_TRUTH: GROUND_TRUTH_VALUES,
                                          'ground_truth_timestamp': GROUND_TRUTH_TIMESTAMP_VALUES,
                                          'partner_inference_id': PARTNER_INFERENCE_ID_VALUES}),
            partner_inf_ids=None, expected_result=GT_ONLY_DF), id="gt-data-only-with-ts"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame({GROUND_TRUTH: GROUND_TRUTH_VALUES}), partner_inf_ids=None,
            expected_result=UserValueError), id="gt-data-only-missing-ids"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame(INFERENCE_DATA), partner_inf_ids=PARTNER_INFERENCE_ID_VALUES,
            expected_result=INFERENCE_ONLY_DF, ignore_join_errors=True), id="inf-data-no-metadata"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame(INFERENCE_DATA), partner_inf_ids=None,
            expected_result=UserValueError), id="inf-data-no-metadata-error"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame({**INFERENCE_DATA, 'partner_inference_id': PARTNER_INFERENCE_ID_VALUES}),
            partner_inf_ids=None, expected_result=INFERENCE_ONLY_DF), id="inf-data-partner-ids-only"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame({**INFERENCE_DATA, 'inference_timestamp': INFERENCE_TIMESTAMP_VALUES}),
            partner_inf_ids=PARTNER_INFERENCE_ID_VALUES, expected_result=INFERENCE_ONLY_DF, ignore_join_errors=True),
            id="inf-data-ts-only"),
        pytest.param(BulkInferencesCase(
            input_dataframe=pd.DataFrame({**INFERENCE_DATA, 'inference_timestamp': INFERENCE_TIMESTAMP_VALUES}),
            partner_inf_ids=None, expected_result=UserValueError),
            id="inf-data-ts-only-error")
    ]

    @responses.activate
    @pytest.mark.parametrize("case", bulk_inferences_metadata_cases)
    def test_send_bulk_inferences_missing_metadata(self, tabular_model: ArthurModel, case: BulkInferencesCase):
        self._mock_batch_upload(TABULAR_MODEL_ID, 'batch_1')
        uuid_generator = MockShortUUID(case.partner_inf_ids)
        datetime_generator = MockDatetime([self.TIMESTAMP_VALUE], return_raw_strings=True)

        with patch("arthurai.core.inferences.shortuuid.uuid") as uuid_mock:
            with patch("arthurai.core.inferences.datetime") as datetime_mock:
                uuid_mock.side_effect = uuid_generator.next
                datetime_mock.now.side_effect = datetime_generator.next
                kwargs = {'data': case.input_dataframe, 'batch_id': 'batch_1'}
                if case.ignore_join_errors is not None:
                    kwargs['ignore_join_errors'] = case.ignore_join_errors
                if isinstance(case.expected_result, pd.DataFrame):
                    resp, resp_close = tabular_model.send_bulk_inferences(**kwargs)
                else:
                    with pytest.raises(case.expected_result):
                        tabular_model.send_bulk_inferences(**kwargs)

        if isinstance(case.expected_result, pd.DataFrame):
            self._assert_result(resp)
            assert resp_close == {"dataset_close_result": "success"}

        assert uuid_generator.call_count() == (len(case.partner_inf_ids) if case.partner_inf_ids is not None else 0)

    @pytest.mark.parametrize("case", bulk_inferences_metadata_cases)
    def test_add_inference_metadata_to_dataframe(self, tabular_model: ArthurModel, case: BulkInferencesCase):
        uuid_generator = MockShortUUID(case.partner_inf_ids)
        datetime_generator = MockDatetime([self.TIMESTAMP_VALUE], return_raw_strings=True)

        input_df_copy = case.input_dataframe.copy(deep=True)

        with patch("arthurai.core.inferences.shortuuid.uuid") as uuid_mock:
            with patch("arthurai.core.inferences.datetime") as datetime_mock:
                uuid_mock.side_effect = uuid_generator.next
                datetime_mock.now.side_effect = datetime_generator.next
                kwargs = {'df': case.input_dataframe, 'model_attributes': tabular_model.attributes}
                if case.ignore_join_errors is not None:
                    kwargs['ignore_join_errors'] = case.ignore_join_errors
                if isinstance(case.expected_result, pd.DataFrame):
                    actual_output_df = add_inference_metadata_to_dataframe(**kwargs)
                else:
                    with pytest.raises(case.expected_result):
                        add_inference_metadata_to_dataframe(**kwargs)

        if isinstance(case.expected_result, pd.DataFrame):
            # don't care about column order so sort before comparing
            pd.testing.assert_frame_equal(actual_output_df.reindex(sorted(actual_output_df.columns), axis=1),
                                          case.expected_result.reindex(sorted(case.expected_result.columns), axis=1))

        assert uuid_generator.call_count() == (len(case.partner_inf_ids) if case.partner_inf_ids is not None else 0)
        # assert input dataframe not modified
        pd.testing.assert_frame_equal(case.input_dataframe, input_df_copy)

    @staticmethod
    def _assert_result(resp: Any, success_count: int = 100, failure_count: int = 0, total_count: int = 100):
        assert resp["counts"]["success"] == success_count
        assert resp["counts"]["failure"] == failure_count
        assert resp["counts"]["total"] == total_count



import json
from datetime import datetime
from unittest.mock import MagicMock

import pytz
import responses

from tests.fixtures.mocks import BASE_URL


CONTENT_TYPE_JSON = 'application/json'


def mock_post(append_url, dict_to_return, status=200):
    """
    Adds a mock 200 response

    :param append_url: the url to the base url
    :param dict_to_return: the dictionary to return as json
    :return: None
    """

    responses.add(responses.POST, BASE_URL + append_url,
                  json.dumps(dict_to_return), status=status,
                  content_type=CONTENT_TYPE_JSON)


def mock_patch(append_url, dict_to_return, status=200):
    """
    Adds a mock 200 response

    :param append_url: the url to the base url
    :param dict_to_return: the dictionary to return as json
    :return: None
    """

    responses.add(responses.PATCH, BASE_URL + append_url,
                  json.dumps(dict_to_return), status=status,
                  content_type=CONTENT_TYPE_JSON)


def mock_get(append_url, response_body, request_type=responses.GET, status=200, content_type=CONTENT_TYPE_JSON):
    """
    Adds a mock 200 response

    :param append_url: the url to the base url
    :param response_content: the response_content to return from the mock API request
    :return: None
    """

    responses.add(request_type,
                  BASE_URL + append_url,
                  body=response_body,
                  status=status,
                  content_type=content_type)


class MockShortUUID:
    def __init__(self, ids=None):
        if ids is None:
            self.ids = []
        else:
            self.ids = ids
        self.next_idx = 0

    def call_count(self):
        return self.next_idx

    def next(self):
        value = self.ids[self.next_idx]
        self.next_idx += 1
        return value


class MockDatetime:
    def __init__(self, return_vals=None, return_raw_strings=False):
        if return_vals is None or len(return_vals) == 0:
            self.timestamps = []
        else:
            if isinstance(return_vals[0], datetime):
                self.timestamps = return_vals
            elif isinstance(return_vals[0], str):
                mocks = []
                for ts in return_vals:
                    if return_raw_strings:
                        mocks.append(ts)
                    else:
                        dt = MagicMock(datetime)
                        dt.isoformat.return_value = ts
                        mocks.append(dt)
                self.timestamps = mocks
        self.next_idx = 0

    def call_count(self):
        return self.next_idx

    def next(self, tz):
        if tz != pytz.utc:
            raise ValueError("unexpected timezone")
        value = self.timestamps[self.next_idx]
        self.next_idx += 1
        return value
